@@ -1,11 +1,13 @@
 
-from __future__ import absolute_import, print_function
-import errno, os, stat, struct, tempfile
+from contextlib import ExitStack
+import errno, os, stat, struct
 
-from bup import compat, metadata, xstat
+from bup import metadata, xstat
 from bup._helpers import UINT_MAX, bytescmp
-from bup.compat import range
-from bup.helpers import (add_error, log, merge_iter, mmap_readwrite,
+from bup.compat import pending_raise
+from bup.helpers import (add_error,
+                         atomically_replaced_file,
+                         log, merge_iter, mmap_readwrite,
                          progress, qprogress, resolve_parent, slashappend)
 
 EMPTY_SHA = b'\0' * 20
@@ -54,12 +56,19 @@ class MetaStoreReader:
         self._file = open(filename, 'rb')
 
     def close(self):
-        if self._file:
-            self._file.close()
-            self._file = None
+        f, self._file = self._file, None
+        if f:
+            f.close()
 
     def __del__(self):
-        self.close()
+        assert not self._file
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, traceback):
+        with pending_raise(value, rethrow=True):
+            self.close()
 
     def metadata_at(self, ofs):
         self._file.seek(ofs)
@@ -71,6 +80,7 @@ class MetaStoreWriter:
     # truncation or corruption somewhat sensibly.
 
     def __init__(self, filename):
+        self._closed = False
         # Map metadata hashes to bupindex.meta offsets.
         self._offsets = {}
         self._filename = filename
@@ -97,13 +107,20 @@ class MetaStoreWriter:
         self._file = open(filename, 'ab')
 
     def close(self):
+        self._closed = True
         if self._file:
             self._file.close()
             self._file = None
 
     def __del__(self):
-        # Be optimistic.
-        self.close()
+        assert self._closed
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, traceback):
+        with pending_raise(value, rethrow=False):
+            self.close()
 
     def store(self, metadata):
         meta_encoded = metadata.encode(include_path=False)
@@ -127,7 +144,7 @@ class Level:
         (ofs,n) = (f.tell(), len(self.list))
         if self.list:
             count = len(self.list)
-            #log('popping %r with %d entries\n' 
+            #log('popping %r with %d entries\n'
             #    % (''.join(self.ename), count))
             for e in self.list:
                 e.write(f)
@@ -165,8 +182,8 @@ def _golevel(level, f, ename, newentry, metastore, tmax):
 
 class Entry:
     def __init__(self, basename, name, meta_ofs, tmax):
-        assert basename is None or type(basename) == bytes
-        assert name is None or type(name) == bytes
+        assert basename is None or isinstance(basename, bytes)
+        assert name is None or isinstance(name, bytes)
         self.basename = basename
         self.name = name
         self.meta_ofs = meta_ofs
@@ -402,10 +419,11 @@ class ExistingEntry(Entry):
 
     def __iter__(self):
         return self.iter()
-            
+
 
 class Reader:
     def __init__(self, filename):
+        self.closed = False
         self.filename = filename
         self.m = b''
         self.writable = False
@@ -432,8 +450,12 @@ class Reader:
                                                self.m[st.st_size - FOOTLEN
                                                       : st.st_size])[0]
 
-    def __del__(self):
-        self.close()
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, traceback):
+        with pending_raise(value, rethrow=False):
+            self.close()
 
     def __len__(self):
         return int(self.count)
@@ -477,12 +499,16 @@ class Reader:
             self.m.flush()
 
     def close(self):
+        self.closed = True
         self.save()
         if self.writable and self.m:
             self.m.close()
             self.m = None
             self.writable = False
 
+    def __del__(self):
+        assert self.closed
+
     def filter(self, prefixes, wantrecurse=None):
         for (rp, path) in reduce_paths(prefixes):
             any_entries = False
@@ -496,7 +522,8 @@ class Reader:
                 # Otherwise something like "save x/y" will produce
                 # nothing if x is up to date.
                 pe = self.find(rp)
-                assert(pe)
+                if not pe:
+                    raise Exception("cannot find %r" % rp)
                 name = path + pe.name[len(rp):]
                 yield (name, pe)
 
@@ -513,7 +540,9 @@ def pathsplit(p):
 
 class Writer:
     def __init__(self, filename, metastore, tmax):
+        self.closed = False
         self.rootlevel = self.level = Level([], None)
+        self.pending_index = None
         self.f = None
         self.count = 0
         self.lastfile = None
@@ -522,19 +551,24 @@ class Writer:
         self.metastore = metastore
         self.tmax = tmax
         (dir,name) = os.path.split(filename)
-        ffd, self.tmpname = tempfile.mkstemp(b'.tmp', filename, dir)
-        self.f = os.fdopen(ffd, 'wb', 65536)
-        self.f.write(INDEX_HDR)
+        with ExitStack() as self.cleanup:
+            self.pending_index = atomically_replaced_file(self.filename,
+                                                          mode='wb',
+                                                          buffering=65536)
+            self.f = self.cleanup.enter_context(self.pending_index)
+            self.cleanup.enter_context(self.f)
+            self.f.write(INDEX_HDR)
+            self.cleanup = self.cleanup.pop_all()
 
-    def __del__(self):
-        self.abort()
+    def __enter__(self):
+        return self
+
+    def __exit__(self, type, value, traceback):
+        with pending_raise(value, rethrow=False):
+            self.abort()
 
     def abort(self):
-        f = self.f
-        self.f = None
-        if f:
-            f.close()
-            os.unlink(self.tmpname)
+        self.close(abort=True)
 
     def flush(self):
         if self.level:
@@ -547,17 +581,20 @@ class Writer:
             self.f.flush()
         assert(self.level == None)
 
-    def close(self):
-        self.flush()
-        f = self.f
-        self.f = None
-        if f:
-            f.close()
-            os.rename(self.tmpname, self.filename)
+    def close(self, abort=False):
+        self.closed = True
+        with self.cleanup:
+            if abort:
+                self.pending_index.cancel()
+            else:
+                self.flush()
+
+    def __del__(self):
+        assert self.closed
 
     def _add(self, ename, entry):
         if self.lastfile and self.lastfile <= ename:
-            raise Error('%r must come before %r' 
+            raise Error('%r must come before %r'
                              % (''.join(ename), ''.join(self.lastfile)))
         self.lastfile = ename
         self.level = _golevel(self.level, self.f, ename, entry,
@@ -598,7 +635,7 @@ class Writer:
 
     def new_reader(self):
         self.flush()
-        return Reader(self.tmpname)
+        return Reader(self.f.name)
 
 
 def _slashappend_or_add_error(p, caller):
@@ -634,7 +671,7 @@ def reduce_paths(paths):
     paths = []
     prev = None
     for (rp, p) in xpaths:
-        if prev and (prev == rp 
+        if prev and (prev == rp
                      or (prev.endswith(b'/') and rp.startswith(prev))):
             continue # already superceded by previous path
         paths.append((rp, p))
